# -*- coding: utf-8 -*-
"""
vectorized pytorch implementations of free space Green's tensors
"""
# %%
import warnings

import numpy as np
import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX

from torchgdm.env.base_classes import EnvironmentBase
from torchgdm.tools.misc import tensor_is_diagonal_and_identical, test_positional_input
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.special import H1n


# %%
# --- vectorized pytorch
def G0_2d_Ep(
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    eps_env: float,
    k0_y: float,
    epsilon_dist: float = 1e-8,
) -> torch.Tensor:
    """2D free space electric field (at r_probe) Green's tensor for an electric line dipole (at r_source)

    Args:
        r_probe (torch.Tensor): probe position
        r_source (torch.Tensor): source position
        wavelength (float): in nm
        eps_env (float): environment permittivity
        k0_y (float): parallel wavevector component (must match the illumination)
        epsilon_dist (float, optional): small distance to avoid divergence at r_probe=r_source. Defaults to 1e-8.

    Returns:
        torch.Tensor: 2D electric-electric Green's Dyad
    """
    # note: hankel support only pure real / pure imag args
    eps_env = torch.as_tensor(eps_env, dtype=DTYPE_COMPLEX, device=r_probe.device).real

    r_probe, r_source = test_positional_input(r_probe, r_source)

    # positional values
    delta_r = r_probe - r_source  # vector difference(s)
    d_x = delta_r[..., 0]
    d_z = delta_r[..., 2]
    y_obs = r_probe[..., 1]
    lR = torch.sqrt(d_x**2 + d_z**2 + epsilon_dist)

    # wavenumber
    k0 = 2 * np.pi / wavelength
    k02 = k0**2
    k0y2 = k0_y**2
    kr2 = eps_env * k02 - k0y2
    kr = torch.sqrt(kr2)  # wavenumber in 2D plane

    # evaluate Hankel functions of first kind
    h01 = H1n(0, kr * lR)
    h11 = H1n(1, kr * lR)
    h21 = H1n(2, kr * lR)

    # pre-calculate termes used multiple times
    cop = d_x / lR
    sip = d_z / lR
    co2p = 2.0 * cop**2 - 1.0
    si2p = 2.0 * cop * sip
    xx_t1 = (k02 - kr2 / (2.0 * eps_env)) * h01
    xx_t2 = kr2 * co2p * h21 / (2.0 * eps_env)

    # phase along infinite axis
    phase_y = torch.exp(1j * k0_y * y_obs)

    # populate the Green's tensor
    G0_2d = torch.zeros(
        (*delta_r.size()[:-1], 3, 3), dtype=DTYPE_COMPLEX, device=r_probe.device
    )
    G0_2d[..., 0, 0] = (xx_t1 + xx_t2) * phase_y
    G0_2d[..., 0, 1] = -1j * kr * k0_y * cop * h11 / eps_env * phase_y
    G0_2d[..., 0, 2] = (kr2 * si2p) * h21 / (2.0 * eps_env) * phase_y
    G0_2d[..., 1, 1] = (k02 - k0y2 / eps_env) * h01 * phase_y
    G0_2d[..., 1, 2] = -1j * kr * k0_y * sip * h11 / eps_env * phase_y
    G0_2d[..., 2, 2] = (xx_t1 - xx_t2) * phase_y

    G0_2d[..., 1, 0] = G0_2d[..., 0, 1]
    G0_2d[..., 2, 0] = G0_2d[..., 0, 2]
    G0_2d[..., 2, 1] = G0_2d[..., 1, 2]

    # here: cgs! SI: *1/(4*pi)
    return 1j * torch.pi * G0_2d


def G0_2d_Hp(
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    eps_env: float,
    k0_y: float,
    epsilon_dist: float = 1e-8,  # to avoid divergence at r_probe=r_source
) -> torch.Tensor:
    """2D free space Green's tensor for magnetic field (at r_probe) from an electric line dipole (at r_source)

    Args:
        r_probe (torch.Tensor): probe position
        r_source (torch.Tensor): source position
        wavelength (float): in nm
        eps_env (float): environment permittivity
        k0_y (float): parallel wavevector component (must match the illumination)
        epsilon_dist (float, optional): small distance to avoid divergence at r_probe=r_source. Defaults to 1e-8.

    Returns:
        torch.Tensor: 2D magnetic-magnetic Green's Dyad
    """
    eps_env = torch.as_tensor(eps_env, dtype=DTYPE_COMPLEX, device=r_probe.device).real

    r_probe, r_source = test_positional_input(r_probe, r_source)

    # positional values
    delta_r = r_probe - r_source  # vector difference(s)
    d_x = delta_r[..., 0]
    d_z = delta_r[..., 2]
    y_obs = r_probe[..., 1]
    lR = torch.sqrt(d_x**2 + d_z**2 + epsilon_dist)

    # wavenumber
    k0 = 2 * np.pi / wavelength
    k02 = k0**2
    k0y2 = k0_y**2
    kr2 = eps_env * k02 - k0y2
    kr = torch.sqrt(kr2)  # wavenumber in 2D plane

    # evaluate Hankel functions of first kind
    if k0_y != 0:
        h01 = H1n(0, kr * lR)
    else:
        h01 = 0
    h11 = H1n(1, kr * lR)

    # phase along infinite axis
    phase_y = torch.exp(1j * k0_y * y_obs)

    # populate the Green's tensor
    G0_2d = torch.zeros(
        (*delta_r.size()[:-1], 3, 3), dtype=DTYPE_COMPLEX, device=r_probe.device
    )
    G0_2d[..., 0, 0] = 0
    G0_2d[..., 0, 1] = k0 * torch.pi * h11 * (kr / lR) * d_z * phase_y
    G0_2d[..., 0, 2] = 1j * torch.pi * k0 * k0_y * h01 * phase_y

    G0_2d[..., 1, 0] = -1 * G0_2d[..., 0, 1]
    G0_2d[..., 1, 1] = 0
    G0_2d[..., 1, 2] = k0 * torch.pi * h11 * (kr / lR) * d_x * phase_y

    G0_2d[..., 2, 0] = -1 * G0_2d[..., 0, 2]
    G0_2d[..., 2, 1] = -1 * G0_2d[..., 1, 2]
    G0_2d[..., 2, 2] = 0

    # here: cgs! SI: *1/(4*pi)
    return G0_2d


def G0_2d_Em(
    r_probe, r_source, wavelength, eps_env, k0_y=0.0, epsilon_dist: float = 1e-8
):
    """2D free space electric field (at r_probe) Green's tensor for a magnetic line dipole (at r_source)

    G0_2d_Em = - G0_2d_Hp

    for doc, see :func:`G0_2d_Hp`
    """
    return -1 * G0_2d_Hp(
        r_probe,
        r_source,
        wavelength,
        eps_env=eps_env,
        k0_y=k0_y,
        epsilon_dist=epsilon_dist,
    )


def G0_2d_Hm(
    r_probe, r_source, wavelength, eps_env, k0_y=0.0, epsilon_dist: float = 1e-8
):
    """2D free space magnetic field (at r_probe) Green's tensor for a magnetic line dipole (at r_source)

    G0_2d_Hm = G0_2d_Ep

    for doc, see :func:`G0_2d_Ep`
    """
    return eps_env * G0_2d_Ep(
        r_probe,
        r_source,
        wavelength,
        eps_env=eps_env,
        k0_y=k0_y,
        epsilon_dist=epsilon_dist,
    )


# --- environment class
class EnvHomogeneous2D(EnvironmentBase):
    """class defining a homogeneous 2D environment

    - defines set of free-space Green's tensors
    - environemnt material needs to be isotropic (scalar epsilon) and lossless.
    - infinite axis is the y-axis
    """

    __name__ = "homogeneous environment 2D"

    def __init__(
        self,
        env_material=1.0,
        k0_y=0.0,
        inc_angle_y=None,
        device: torch.device = None,
    ):
        """class defining a homogeneous 2D environment. Infinite axis along y

        Args:
            env_material (float, optional): Environment material. Either float or class from :mod:`torchgdm.materials`. A float value is interpreted as permittivity.  Defaults to 1.0.
            k0_y (float, optional): parallel k-vector component. Must match the illumination field. Caution, using a `k0_y != 0` will be constant for all wavelengths! Defaults to 0.0 (normal incidence).
            inc_angle_y (float, optional): If used, `k0_y` will be ignored. Incident angle to calculate parallel k-vector component. Must match the illumination field.  Using the incident angle instead of `k0_y` will work for several wavelengths. Defaults to `None` (--> use `k0_y`).
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)
        self.n_dim = 2

        if type(env_material) in (float, int):
            from torchgdm.materials import MatConstant

            self.env_material = MatConstant(env_material)
        else:
            self.env_material = env_material
        

        # set parallel k-vector comp.
        self.k0_y = torch.as_tensor(k0_y, dtype=DTYPE_FLOAT, device=device)
        self.inc_angle_y = inc_angle_y
        if inc_angle_y is not None:
            self.inc_angle_y = torch.as_tensor(
                inc_angle_y, dtype=DTYPE_FLOAT, device=self.device
            )
        
        self.set_device(self.device)

    def __repr__(self, verbose=False):
        """description about simulation environment defined by set of dyads"""
        out_str = " ------ homogeneous 2D environment, infinite axis along Y -------"
        out_str += "\n env. material: {}".format(self.env_material)
        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

        if self.inc_angle_y is not None:
            self.inc_angle_y = self.inc_angle_y.to(device)

        self.k0_y = self.k0_y.to(device)

        self.env_material.set_device(device)

    def get_k0_y(self, wavelength):
        """return parallel k-vector component"""
        if self.inc_angle_y is not None:
            return torch.as_tensor(
                torch.sin(self.inc_angle_y) * 2 * torch.pi / wavelength,
                dtype=DTYPE_FLOAT,
                device=self.device,
            )
        else:
            return self.k0_y

    def get_environment_permittivity_scalar(
        self, wavelength: float, r_probe: torch.Tensor
    ):
        """return scalar complex environment permittivity for `wavelength` at pos. `r_probe` (units in nm)

        assume isotropic env. material: returns [0,0]-component of eps. tensor

        homogeneous medium: all positions `r_probe` will be same material

        Args:
            wavelength (float): in nm
            r_probe (torch.Tensor): position(s) at which the permittivity is to be returned

        Raises:
            ValueError: In case env. material is not isotropic

        Returns:
            torch.Tensor: scalar permittivity at all positions `r_probe`
        """

        eps_env_tensor = self.env_material.get_epsilon(wavelength)

        if not tensor_is_diagonal_and_identical(eps_env_tensor):
            raise ValueError(
                "Environment material is not isotropic, but tensorial epsilon was received. "
                + "Only isotropic environment media are supported."
            )
        self.eps_env_scalar = eps_env_tensor[0, 0]

        if self.eps_env_scalar.imag != 0:
            warnings.warn(
                "Environment permittivity evaluated to non-zero imaginary part. "
                + "This is not physically meaningful in the infinite environment medium."
            )

        eps_env_tensor = (
            torch.ones(len(r_probe), dtype=DTYPE_COMPLEX, device=self.device)
            * self.eps_env_scalar
        )

        return eps_env_tensor

    # --- individual sources Green's tensors
    def get_G_Ep(self, r_probe, r_source, wavelength):
        """2D free space electric field (at r_probe) Green's tensor for an electric line dipole (at r_source)

        Args:
            r_probe (torch.Tensor): probe position
            r_source (torch.Tensor): source position
            wavelength (float): in nm

        Returns:
            torch.Tensor: 2D electric-electric Green's Dyad
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        k0_y = self.get_k0_y(wavelength)
        return G0_2d_Ep(r_probe, r_source, wavelength, k0_y=k0_y, eps_env=eps_env)

    def get_G_Hp(self, r_probe, r_source, wavelength):
        """2D free space magnetic field (at r_probe) Green's tensor for a magnetic line dipole (at r_source)

        Args:
            r_probe (torch.Tensor): probe position
            r_source (torch.Tensor): source position
            wavelength (float): in nm

        Returns:
            torch.Tensor: 2D magnetic-magnetic Green's Dyad
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        k0_y = self.get_k0_y(wavelength)
        return G0_2d_Hp(r_probe, r_source, wavelength, k0_y=k0_y, eps_env=eps_env)

    def get_G_Em(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """2D free space electric field (at r_probe) Green's tensor for a magnetic line dipole (at r_source)

        G0_2d_Em = - G0_2d_Hp

        for doc, see :meth:`EnvHomogeneous2D.get_G_Hp`
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        k0_y = self.get_k0_y(wavelength)
        return G0_2d_Em(r_probe, r_source, wavelength, k0_y=k0_y, eps_env=eps_env)

    def get_G_Hm(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """2D free space magnetic field (at r_probe) Green's tensor for a magnetic line dipole (at r_source)

        G0_2d_Hm = G0_2d_Ep

        for doc, see :meth:`EnvHomogeneous2D.get_G_Ep`
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        k0_y = self.get_k0_y(wavelength)
        return G0_2d_Hm(r_probe, r_source, wavelength, k0_y=k0_y, eps_env=eps_env)

    def get_G_Ep_farfield(self, r_probe, r_source, wavelength):
        """2D far-field approximation  for electric-electric Green's tensor

        2D case: No approximation is used, return full 2D tensor

        Args:
            r_probe (torch.Tensor): probe position
            r_source (torch.Tensor): source position (electric line dipole)
            wavelength (float): in nm

        Returns:
            torch.Tensor: 2D farfield electric-electric Green's Dyad
        """
        return self.get_G_Ep(r_probe, r_source, wavelength)

    def get_G_Em_farfield(self, r_probe, r_source, wavelength):
        """2D far-field approximation for electric-magnetic Green's tensor

        2D case: No approximation is used, return full 2D tensor

        Args:
            r_probe (torch.Tensor): probe position
            r_source (torch.Tensor): source position (magnetic line dipole)
            wavelength (float): in nm

        Returns:
            torch.Tensor: 2D farfield electric-magnetic Green's Dyad
        """
        return self.get_G_Em(r_probe, r_source, wavelength)


## ---------------- test Green's tensors ----------------
if __name__ == "__main__":
    import time

    import scipy
    import numpy as np
    import matplotlib.pyplot as plt

    from pyGDM2 import tools
    from pyGDM2.propagators.propagators_2D import _s0_2D, _G0_2D, _G0_HE_2D

    # device = torch.device("cuda:0")
    device = torch.device("cpu")

    func_numba = _G0_2D
    func_torch = G0_2d_Ep

    # func_numba = _G0_HE_2D
    # func_torch = G0_2d_Hp

    # ---- quantitative test
    # setup
    eps_env = 1.0
    wl = 500.0
    k0_y = 0.5 * 2 * torch.pi / wl

    DX, NX = 50, 50
    Z0 = 0
    r_probe = tools.generate_NF_map_XZ(1, 10, NX, -2, 5, NX, Z0)
    r0 = np.zeros_like(r_probe, dtype=np.float32)  # fix emitter position
    r_probe_t = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT).to(device)
    r0_t = torch.as_tensor(r0, dtype=DTYPE_FLOAT).to(device)

    # eval numba sequential
    G0_vec1 = np.zeros((NX * NX, 3, 3), dtype=np.complex64)
    print("evaluating {} greens tensors".format(len(r_probe)))
    t1a = time.time()
    for i, (_r0, _r) in enumerate(zip(r0, r_probe)):
        xx, yy, zz, xy, xz, yx, yz, zx, zy = func_numba(_r0, _r, wl, eps_env, ky0=k0_y)
        G0 = np.array(
            [
                [xx, xy, xz],
                [yx, yy, yz],
                [zx, zy, zz],
            ],
            dtype=np.complex64,
        )
        G0_vec1[i] = G0
    t1b = time.time()
    print("time sequential: {:.5f}s".format(t1b - t1a))

    # eval torch
    t2a = time.time()
    G0_vec2t = func_torch(r_probe_t, r0_t, wl, eps_env, k0_y=k0_y)
    t2b = time.time()

    # test evaluate using broadcasting
    # interact_NxN = func_torch(
    #     r_probe_t.unsqueeze(1), r_probe_t.unsqueeze(0), wl, eps_env, k0_y
    # )

    print("time torch (on device {}): {:.5f}s".format(device, t2b - t2a))
    print("speed-up torch: x{:.3f}".format((t1b - t1a) / (t2b - t2a)))

    G0_vec1 = G0_vec1.reshape(NX, NX, 3, 3)
    G0_vec2 = G0_vec2t.detach().cpu().numpy()
    G0_vec2 = G0_vec2.reshape(NX, NX, 3, 3)

    print(np.abs(G0_vec1).mean())
    print((np.abs(G0_vec1 - G0_vec2)).mean())

    i1, i2 = 2, 1
    plt.subplot(321)
    plt.title("numpy")
    plt.imshow(G0_vec1[..., i1, i2].real)
    plt.colorbar()
    plt.subplot(323)
    plt.imshow(G0_vec1[..., i1, i2].imag)
    plt.colorbar()
    plt.subplot(325)
    plt.imshow(np.abs(G0_vec1[..., i1, i2]))
    plt.colorbar()

    plt.subplot(322)
    plt.title("pytorch")
    plt.imshow(G0_vec2[..., i1, i2].real)
    plt.colorbar()
    plt.subplot(324)
    plt.imshow(G0_vec2[..., i1, i2].imag)
    plt.colorbar()
    plt.subplot(326)
    plt.imshow(np.abs(G0_vec2[..., i1, i2]))
    plt.colorbar()

    plt.show()



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestInterpolation(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing plane wave illumination.")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

    def test_interp1d(self):
        from torchgdm.tools.interp import interp1d

        try:
            import numpy as np
        except ImportError:
            import warnings

            warnings.warn("Numpy not installed. Skipping interpolation test.")
            return

        for device in self.devices:
            x_e = torch.linspace(-1.2, 1.2, 500, device=device)
            x_t = torch.linspace(-1, 1, 40, device=device)

            y_t = torch.sin(x_t)
            y_t = torch.rand(len(x_t), dtype=torch.complex64, device=device)

            y_e = interp1d(x_e, x_t, y_t)
            y_e_np = torch.as_tensor(
                np.interp(
                    x_e.detach().cpu().numpy(),
                    x_t.detach().cpu().numpy(),
                    y_t.detach().cpu().numpy(),
                )
            )

            # test if numpy and torch interpolations return equal results
            np.testing.assert_almost_equal(y_e_np.imag, y_e.imag.detach().cpu().numpy(), decimal=6)
            np.testing.assert_almost_equal(y_e_np.real, y_e.real.detach().cpu().numpy(), decimal=6)

            if self.verbose:
                print("  - {}: interpolation test passed.".format(device))


# %%


if __name__ == "__main__":
    print("testing interpolation functions.")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestStructureRotation2D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing structure rotations - 2D...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        self.angles = torch.as_tensor([0, 90, 237.5])

        # --- setup a test case
        # - environment
        mat_env = tg.materials.MatConstant(eps=1.0)
        self.env = tg.env.freespace_2d.EnvHomogeneous2D(env_material=mat_env)

        # - illumination field(s)
        self.wl = torch.as_tensor(620.0)
        self.wavelengths = torch.as_tensor([self.wl])
        self.e_inc_list = [
            tg.env.freespace_2d.PlaneWave(e0p=1, e0s=1, inc_angle=angle)
            for angle in self.angles
        ]

        # - structure
        step = 20.0  # nm
        mat_struct = tg.materials.MatConstant(8 + 0.2j)
        self.struct = tg.struct.StructDiscretizedSquare2D(
            tg.struct.surface_2d.rectangle(5, 2),
            step,
            mat_struct,
            radiative_correction=False,
        )

    def test_rotate_discretized(self):
        for device in self.devices:
            self.struct.set_device(device)

            sim_full = tg.simulation.Simulation(
                structures=[self.struct],
                environment=self.env,
                illumination_fields=[self.e_inc_list[0]],
                wavelengths=self.wavelengths,
                device=device,
            )
            sim_full.run(verbose=False, progress_bar=False)
            cs_full_0 = tg.postproc.crosssect.total(sim_full, self.wl)

            # rotate structure and illumination polarization angle and compare
            for i_alpha, alpha in enumerate(self.angles[1:]):
                _sim_f = tg.simulation.Simulation(
                    structures=[self.struct.rotate(alpha)],
                    environment=self.env,
                    illumination_fields=[self.e_inc_list[i_alpha + 1]],
                    wavelengths=self.wavelengths,
                    device=device,
                )

                # run rotated simulation
                _sim_f.run(verbose=False, progress_bar=False)

                # compare scat / extinct of full and effective pola. simulations
                _cs_f = tg.postproc.crosssect.total(_sim_f, self.wl)

                torch.testing.assert_close(
                    cs_full_0["scs"], _cs_f["scs"], atol=1e-3, rtol=1e-5
                )
                torch.testing.assert_close(
                    cs_full_0["ecs"], _cs_f["ecs"], atol=1e-3, rtol=1e-5
                )

            if self.verbose:
                print(
                    "  - {}: discretized 2D structure rotation test passed.".format(
                        device
                    )
                )


class TestStructureTranslation2D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing strucure copy / combining - 2D...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        self.angles = torch.as_tensor([0, 23.7])

        # --- setup a test case
        # - environment
        mat_env = tg.materials.MatConstant(eps=1.0)
        self.env = tg.env.freespace_2d.EnvHomogeneous2D(env_material=mat_env)

        # - illumination field(s)
        self.wl = torch.as_tensor(700.0)
        self.wavelengths = torch.as_tensor([self.wl])
        self.e_inc_list = [
            tg.env.freespace_2d.PlaneWave(e0p=torch.cos(angle), e0s=torch.sin(angle))
            for angle in self.angles
        ]

        # - structure
        step = 25.0  # nm
        mat_struct = tg.materials.MatConstant(6.5 + 0.1j)
        self.struct = tg.struct.StructDiscretizedSquare2D(
            tg.struct.surface_2d.rectangle(5, 3),
            step,
            mat_struct,
            radiative_correction=False,
        )

        # - combined simulation config
        self.sim_conf = dict(
            environment=self.env,
            illumination_fields=self.e_inc_list,
            wavelengths=self.wavelengths,
        )

    def test_translate_copy(self):
        pos1 = [250, 0, 150]
        pos2 = [0, 0, 50]
        for device in self.devices:
            self.struct.set_device(device)

            sim_1 = tg.simulation.Simulation(
                structures=[self.struct + pos1, self.struct + pos2],
                device=device,
                **self.sim_conf,
            )
            sim_1.run(verbose=False, progress_bar=False)
            cs_full_1 = tg.postproc.crosssect.total(sim_1, self.wl)

            sim_2 = tg.simulation.Simulation(
                structures=[self.struct.copy([pos1, pos2])],
                device=device,
                **self.sim_conf,
            )
            sim_2.run(verbose=False, progress_bar=False)
            cs_full_2 = tg.postproc.crosssect.total(sim_2, self.wl)

            torch.testing.assert_close(
                cs_full_1["scs"], cs_full_2["scs"], atol=1e1, rtol=1e-5
            )
            torch.testing.assert_close(
                cs_full_1["ecs"], cs_full_2["ecs"], atol=1e1, rtol=1e-5
            )

        if self.verbose:
            print(
                "  - {}: discretized 2D structure copy / translate test passed.".format(
                    device
                )
            )

    def test_combine(self):
        pos1 = [250, 0, 150]
        pos2 = [0, 0, 50]
        for device in self.devices:
            self.struct.set_device(device)

            sim_1 = tg.simulation.Simulation(
                structures=[self.struct + pos1],
                device=device,
                **self.sim_conf,
            )
            sim_2 = tg.simulation.Simulation(
                structures=[self.struct + pos2],
                device=device,
                **self.sim_conf,
            )

            # - combined sim / combine struct
            sim_combo1 = tg.simulation.Simulation(
                structures=[self.struct + pos1, self.struct + pos2],
                device=device,
                **self.sim_conf,
            )
            sim_combo2 = sim_1.combine(sim_2)

            # - run and compare results
            sim_combo1.run(verbose=False, progress_bar=False)
            sim_combo2.run(verbose=False, progress_bar=False)

            cs_full_1 = tg.postproc.crosssect.total(sim_combo1, self.wl)
            cs_full_2 = tg.postproc.crosssect.total(sim_combo2, self.wl)

            torch.testing.assert_close(
                cs_full_1["scs"], cs_full_2["scs"], atol=1e1, rtol=1e-5
            )
            torch.testing.assert_close(
                cs_full_1["ecs"], cs_full_2["ecs"], atol=1e1, rtol=1e-5
            )

        if self.verbose:
            print(
                "  - {}: discretized 2D simulation combine test passed.".format(device)
            )


class TestStructureRotation3D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing structure rotations...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        self.angles = torch.as_tensor([0, 90, 237.5])

        # --- setup a test case
        # - environment
        mat_env = tg.materials.MatConstant(eps=1.0)
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wl = torch.as_tensor(700.0)
        self.wavelengths = torch.as_tensor([self.wl])
        self.e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=torch.cos(angle), e0s=torch.sin(angle))
            for angle in self.angles
        ]

        # - structure
        step = 15.0  # nm
        mat_struct = tg.materials.MatDatabase("Au")
        self.struct = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(5, 2, 2),
            step,
            mat_struct,
            radiative_correction=False,
        )

    def test_rotate_discretized(self):
        for device in self.devices:
            self.struct.set_device(device)

            sim_full = tg.simulation.Simulation(
                structures=[self.struct],
                environment=self.env,
                illumination_fields=[self.e_inc_list[0]],
                wavelengths=self.wavelengths,
                device=device,
            )
            sim_full.run(verbose=False, progress_bar=False)
            cs_full_0 = tg.postproc.crosssect.total(sim_full, self.wl)

            # rotate structure and illumination polarization angle and compare
            for i_alpha, alpha in enumerate(self.angles[1:]):
                _sf = self.struct.rotate(alpha)
                _sim_f = tg.simulation.Simulation(
                    structures=[_sf],
                    environment=self.env,
                    illumination_fields=[self.e_inc_list[i_alpha + 1]],
                    wavelengths=self.wavelengths,
                    device=device,
                )

                # run rotated simulation
                _sim_f.run(verbose=False, progress_bar=False)

                # compare scat / extinct of full and effective pola. simulations
                _cs_f = tg.postproc.crosssect.total(_sim_f, self.wl)

                torch.testing.assert_close(
                    cs_full_0["scs"], _cs_f["scs"], atol=1e1, rtol=1e-5
                )
                torch.testing.assert_close(
                    cs_full_0["ecs"], _cs_f["ecs"], atol=1e1, rtol=1e-5
                )

            if self.verbose:
                print(
                    "  - {}: discretized structure rotation test passed.".format(device)
                )

    def test_rotate_polarizability(self):
        for device in self.devices:
            self.struct.set_device(device)

            struct_alpha = self.struct.convert_to_effective_polarizability_pair(
                self.env, self.wavelengths, verbose=False, progress_bar=False
            )
            sim_aeff = tg.simulation.Simulation(
                structures=[self.struct],
                environment=self.env,
                illumination_fields=[self.e_inc_list[0]],
                wavelengths=self.wavelengths,
                device=device,
            )
            sim_aeff.run(verbose=False, progress_bar=False)
            cs_aeff_0 = tg.postproc.crosssect.total(sim_aeff, self.wl)

            # rotate structure and illumination polarization angle and compare
            for i_alpha, alpha in enumerate(self.angles[1:]):
                _sa = struct_alpha.rotate(alpha)
                _sim_a = tg.simulation.Simulation(
                    structures=[_sa],
                    environment=self.env,
                    illumination_fields=[self.e_inc_list[i_alpha + 1]],
                    wavelengths=self.wavelengths,
                    device=device,
                )

                # run rotated simulation
                _sim_a.run(verbose=False, progress_bar=False)

                # compare scat / extinct of full and effective pola. simulations
                _cs_a = tg.postproc.crosssect.total(_sim_a, self.wl)

                torch.testing.assert_close(
                    cs_aeff_0["scs"], _cs_a["scs"], atol=1e3, rtol=1e-2
                )
                torch.testing.assert_close(
                    cs_aeff_0["ecs"], _cs_a["ecs"], atol=1e3, rtol=1e-2
                )

            if self.verbose:
                print(
                    "  - {}: effective polarizability structure rotation test passed.".format(
                        device
                    )
                )


class TestStructureTranslation3D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing strucure copy / combining...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        self.angles = torch.as_tensor([0, 23.7])

        # --- setup a test case
        # - environment
        mat_env = tg.materials.MatConstant(eps=1.0)
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wl = torch.as_tensor(700.0)
        self.wavelengths = torch.as_tensor([self.wl])
        self.e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=torch.cos(angle), e0s=torch.sin(angle))
            for angle in self.angles
        ]

        # - structure
        step = 25.0  # nm
        mat_struct = tg.materials.MatDatabase("Al")
        self.struct = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(3, 2, 2),
            step,
            mat_struct,
            radiative_correction=False,
        )

        self.struct_pola = self.struct.convert_to_effective_polarizability_pair(
            self.env, self.wavelengths, verbose=False, progress_bar=False
        )

        # - combined simulation config
        self.sim_conf = dict(
            environment=self.env,
            illumination_fields=self.e_inc_list,
            wavelengths=self.wavelengths,
        )

    def test_translate_copy(self):
        pos1 = [250, 50, 150]
        pos2 = [0, -50, 50]
        for device in self.devices:
            for struct in [self.struct, self.struct_pola]:
                struct.set_device(device)

                sim_1 = tg.simulation.Simulation(
                    structures=[struct + pos1, struct + pos2],
                    device=device,
                    **self.sim_conf,
                )
                sim_1.run(verbose=False, progress_bar=False)
                cs_full_1 = tg.postproc.crosssect.total(sim_1, self.wl)

                sim_2 = tg.simulation.Simulation(
                    structures=[struct.copy([pos1, pos2])],
                    device=device,
                    **self.sim_conf,
                )
                sim_2.run(verbose=False, progress_bar=False)
                cs_full_2 = tg.postproc.crosssect.total(sim_2, self.wl)

                torch.testing.assert_close(
                    cs_full_1["scs"], cs_full_2["scs"], atol=1e1, rtol=1e-5
                )
                torch.testing.assert_close(
                    cs_full_1["ecs"], cs_full_2["ecs"], atol=1e1, rtol=1e-5
                )

            if self.verbose:
                print("  - {}: structure copy / translate test passed.".format(device))

    def test_combine(self):
        pos1 = [250, 20, 150]
        pos2 = [0, -83, 50]
        for device in self.devices:
            for struct in [self.struct, self.struct_pola]:
                struct.set_device(device)

                sim_1 = tg.simulation.Simulation(
                    structures=[struct + pos1],
                    device=device,
                    **self.sim_conf,
                )
                sim_2 = tg.simulation.Simulation(
                    structures=[struct + pos2],
                    device=device,
                    **self.sim_conf,
                )

                # - combined sim / combine struct
                sim_combo1 = tg.simulation.Simulation(
                    structures=[struct + pos1, struct + pos2],
                    device=device,
                    **self.sim_conf,
                )
                sim_combo2 = sim_1.combine(sim_2)

                # - run and compare results
                sim_combo1.run(verbose=False, progress_bar=False)
                sim_combo2.run(verbose=False, progress_bar=False)

                cs_full_1 = tg.postproc.crosssect.total(sim_combo1, self.wl)
                cs_full_2 = tg.postproc.crosssect.total(sim_combo2, self.wl)

                torch.testing.assert_close(
                    cs_full_1["scs"], cs_full_2["scs"], atol=1e1, rtol=1e-5
                )
                torch.testing.assert_close(
                    cs_full_1["ecs"], cs_full_2["ecs"], atol=1e1, rtol=1e-5
                )

        if self.verbose:
            print(
                "  - {}: discretized 3D simulation combine test passed.".format(device)
            )


# %%
if __name__ == "__main__":
    print("testing structure rotation...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestsimulationClass(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing simulation class API...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(1)

        # - illumination field(s)
        self.wls = torch.linspace(550.0, 650, 2)
        self.wl = self.wls[0]
        self.e_inc = tg.env.freespace_3d.PlaneWave(e0p=1, e0s=1)

        # - structure
        step = 20.0  # nm
        self.struct = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(1, 1, 2), step, tg.materials.MatConstant(8 + 0.2j)
        )

        # - sim
        self.sim = tg.simulation.Simulation(
            structures=[self.struct],
            environment=self.env,
            illumination_fields=[self.e_inc],
            wavelengths=self.wls,
        )

    def test_class_api(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # --- single wavelength evaluations
            r_probe = [[0, 0, 0], [150, 200, 0]]
            nf = self.sim.get_nearfield(self.wl, r_probe, progress_bar=False)
            nf2 = tg.postproc.fields.nf(self.sim, self.wl, r_probe, progress_bar=False)
            for key in ["sca", "tot", "inc"]:
                torch.testing.assert_close(nf[key].efield, nf2[key].efield)
                torch.testing.assert_close(nf[key].hfield, nf2[key].hfield)

            r_probe = [[0, 0, 10000]]
            ff = self.sim.get_farfield(self.wl, r_probe, progress_bar=False)
            ff2 = tg.postproc.fields.ff(self.sim, self.wl, r_probe, progress_bar=False)
            for key in ["sca", "tot", "inc"]:
                torch.testing.assert_close(ff[key].efield, ff2[key].efield)

            # --- spectra
            # - crosssections
            cs = self.sim.get_crosssections(progress_bar=False)
            cs2 = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.crosssect.total, progress_bar=False
            )
            for key in ["wavelengths", "scs", "ecs", "acs"]:
                torch.testing.assert_close(cs[key], cs2[key])

            ecs = self.sim.get_spectra_ecs(progress_bar=False)
            torch.testing.assert_close(ecs["ecs"], cs2["ecs"])

            # - field intensity
            # E-field
            r_probe = [0, 0, 100]
            I_nfe = self.sim.get_spectra_nf_intensity_e(r_probe, progress_bar=False)

            I_nfe2 = tg.tools.batch.calc_spectrum(
                self.sim,
                tg.postproc.fields.integrated_nf_intensity_e,
                r_probe=r_probe,
                progress_bar=False,
            )
            for key in ["wavelengths", "sca", "tot", "inc"]:
                torch.testing.assert_close(I_nfe[key], I_nfe2[key])

            # H-field
            I_nfh = self.sim.get_spectra_nf_intensity_h(r_probe, progress_bar=False)

            I_nfh2 = tg.tools.batch.calc_spectrum(
                self.sim,
                tg.postproc.fields.integrated_nf_intensity_h,
                r_probe=r_probe,
                progress_bar=False,
            )
            for key in ["wavelengths", "sca", "tot", "inc"]:
                torch.testing.assert_close(I_nfh[key], I_nfh2[key])

            # far-field
            r_probe = [0, 0, 10000]
            I_ff = self.sim.get_spectra_ff_intensity(r_probe, progress_bar=False)
            I_ff2 = tg.tools.batch.calc_spectrum(
                self.sim,
                tg.postproc.fields.integrated_ff_intensity,
                r_probe=r_probe,
                progress_bar=False,
            )
            for key in ["wavelengths", "sca", "tot", "inc"]:
                torch.testing.assert_close(I_ff[key], I_ff2[key])

            # multipole decomposition
            mp = self.sim.get_spectra_multipole(progress_bar=False)
            mp2 = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.multipole.decomposition_exact, progress_bar=False
            )
            for key in [
                "ed_1",
                "ed_toroidal",
                "ed_tot",
                "md",
                "eq_1",
                "eq_toroidal",
                "eq_tot",
                "mq",
                "wavelengths",
            ]:
                torch.testing.assert_close(mp[key], mp2[key])

            mp_scs = self.sim.get_spectra_multipole_scs(progress_bar=False)
            mp_scs2 = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.multipole.scs, progress_bar=False
            )
            for key in ["scs_ed", "scs_md", "scs_eq", "scs_mq", "wavelengths"]:
                torch.testing.assert_close(mp_scs[key], mp_scs2[key])

            mp_ecs = self.sim.get_spectra_multipole_ecs(progress_bar=False)
            mp_ecs2 = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.multipole.ecs, progress_bar=False
            )
            for key in ["ecs_ed", "ecs_md", "ecs_eq", "ecs_mq", "wavelengths"]:
                torch.testing.assert_close(mp_ecs[key], mp_ecs2[key])

            if self.verbose:
                print("  - {}: simulation class API test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing simulation class API...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""
core simulation class
"""
# %%
import warnings
import time

import torch

from torchgdm.field import Field
from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import tqdm, purge_mem, get_default_device
from torchgdm.tools.batch import get_e0, get_h0
from torchgdm.tools.geometry import test_structure_distances


# --- simulation base class
class SimulationBase:
    """torchgdm simulation container

    This class contains all elements describing a simulation:
     - the environment
     - the structure(s)
     - the illumination field(s)
     - the evaluation wavelength(s)

    It also acts as an interface to running and postprocessing
    of the simulation as well as for several visualization tools.
     - run the simulation
     - evaluate observables
     - evaluate spectra
     - evaluate a rasterscan

    It also contains methods for manipulations of the simulation and structures
     - add / remove a structure
     - split off a new simulation with a sub-structure
     - combine with structures and their fields from another simulation
     - copy (including optional shift of all structures)

    Furthermore it several contains private methods necessary for simulation setup and running
    """

    __name__ = "simulation base class"

    def __init__(
        self,
        structures: list,
        environment,
        illumination_fields: list,
        wavelengths: torch.Tensor,
        test_structure_collisions: bool = True,
        on_distance_violation: str = "error",
        device: torch.device = None,
    ):
        """base simulation constructor

        Args:
            structures (_type_): list of structures
            environment (_type_): environment class
            illumination_fields (_type_): list of illuminations
            wavelengths (torch.Tensor): evaluation wavelengths (in nm)
            test_structure_collisions (bool, optional): Check for overlapping / colliding structures. Can be slow in case of many structures. Defaults to True.
            on_distance_violation (str, optional): behavior on distance violation. can be "error", "warn" or None (do nothing). Defaults to "error".
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # structures, wavelengths, illuminations: if single element given, wrap into list
        if hasattr(structures, "__iter__"):
            self.structures = [
                _s for _s in structures if _s is not None
            ]  # ignore "None" entries
        else:
            self.structures = [structures]

        if hasattr(illumination_fields, "__iter__"):
            self.illumination_fields = illumination_fields
        else:
            self.illumination_fields = [illumination_fields]

        # convert single wavelength to list
        self.wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        self.wavelengths = torch.atleast_1d(self.wavelengths)

        
        # environment
        if environment is None:
            from torchgdm.env import EnvHomogeneous3D

            env = EnvHomogeneous3D(device=device)
            warnings.warn("No environment specified. Using a vacuum 3D environment.")
        elif type(environment) in [float, int]:
            from torchgdm.env import EnvHomogeneous3D

            env = EnvHomogeneous3D(env_material=environment, device=device)
            warnings.warn(
                "Environment is a scalar, use 3D environment with medium permittivity eps={}.".format(
                    environment
                )
            )
        else:
            env = environment
        self.environment = env

        # test collisions:
        for _s1 in self.structures:
            for _s2 in self.structures:
                if _s1 != _s2:
                    N_dist1, N_dist2 = test_structure_distances(
                        _s1, _s2, on_distance_violation=on_distance_violation
                    )

        self.reset_fields()

        # test if all components are same dimension as environment
        self._test_dimensions_matching()

    def _test_dimensions_matching(self):
        self.n_dim = dim = self.environment.n_dim
        if dim == -1:
            raise ValueError(
                "The environemnt has undefined dimension (-1), please modify your environment class."
            )

        # test structures
        for _s in self.structures:
            if _s.n_dim == -1:
                raise ValueError(
                    "A structure has undefined dimension (-1), please modify your structure class. "
                    + f"The current environment is {dim}D."
                )
            if _s.n_dim != dim:
                raise ValueError(
                    "Structures with inconsistent dimensions found. "
                    + f"Environment is {dim}D, all structures must be {dim}D as well."
                )

        # test illumination fields
        for _f in self.illumination_fields:
            if _f.n_dim == -1:
                raise ValueError(
                    "An illumination field has undefined dimension (-1), please modify your field class. "
                    + f"The current environment is {dim}D."
                )
            if _f.n_dim != dim:
                raise ValueError(
                    "Illumination fields with inconsistent dimensions found. "
                    + f"Environment is {dim}D, all fields must be {dim}D as well."
                )

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device

        # iterate lists of torchgdm objects
        for i in range(len(self.structures)):
            self.structures[i].set_device(device=device)
        for i in range(len(self.illumination_fields)):
            self.illumination_fields[i].set_device(device=device)

        # stored fields
        for k in self.fields_inside:
            self.fields_inside[k].set_device(device=device)

        # further objects
        self.environment.set_device(device=device)
        self.wavelengths = self.wavelengths.to(device=device)

    # --- simulation manipulation
    def add_struct(self, struct, on_distance_violation="error"):
        """add a structure to the simulation

        reset possibly pre-calculated fields.
        **Caution:** Not compatible with autograd!

        Args:
            struct (StructBase): structure to add

        Returns:
            simulation including added structure
        """
        import copy

        _new_sim = copy.deepcopy(self)

        _new_sim.structures.append(struct)

        N_dist1, N_dist2 = test_structure_distances(
            self, struct, on_distance_violation=on_distance_violation
        )

        warnings.warn(
            "Resetting pre-calculated fields. This should be changed, e.g. add zero fields."
        )
        _new_sim.reset_fields()
        return _new_sim

    def delete_struct(self, index: int):
        """delete a structure from the simulation

        **Caution:** Not compatible with autograd!

        Args:
            index (int): index of structure to delete from sim

        Returns:
            simulation without deleted structure
        """
        _simlist = [self.split(i) for i in range(len(self.structures))]
        if index == 0:
            _s = _simlist.pop(0)

        _new_sim = _simlist.pop(0)
        for i, _s in enumerate(_simlist):
            if index != 0 and i == index:
                pass  # delete (don't add) this structure
            else:
                _new_sim.combine(_s)

        return _new_sim

    def split(self, structure_index: int):
        """split a structure into new simulation

        return a new simulation with identical config,
        containing only the splitted structure.
        If fields were calculated, they are kept for the splitted structure

        **Caution:** Not compatible with autograd!

        Args:
            index (int): index of structure to split from sim

        Returns:
            simulation containing splitted structure
        """
        import copy

        _new_sim = copy.deepcopy(self)

        # if calculated, extract fields inside splitted structure
        if len(_new_sim.fields_inside) > 0:
            _new_sim.fields_inside = _new_sim._get_insidefields_for_single_structure(
                structure_index=structure_index
            )

        # retain selected structure
        _new_sim.structures = [_new_sim.structures[structure_index]]

        return _new_sim

    def combine(self, other_sim, on_distance_violation="error"):
        """combine with other simulation

        Combining pre-calculated fields. Wavelengths and field configs must match!

        **Caution:** Not compatible with autograd!

        Args:
            other_sim (:class:`Simulation`): simulation to combine with

        Returns:
            :class:`Simulation`: combined simulation
        """
        import copy

        _new_sim = copy.deepcopy(self)

        assert len(_new_sim.fields_inside) == len(other_sim.fields_inside)
        assert _new_sim.wavelengths == other_sim.wavelengths
        assert len(_new_sim.illumination_fields) == len(other_sim.illumination_fields)
        assert _new_sim.environment.__name__ == other_sim.environment.__name__
        assert _new_sim.device == other_sim.device

        N_dist1, N_dist2 = test_structure_distances(
            self, other_sim, on_distance_violation=on_distance_violation
        )

        # if existing: combine inside fields
        if len(_new_sim.fields_inside) > 0:
            fields_inside_combined = dict()
            for wl in self.wavelengths:
                wl = float(wl)

                e_in_1 = _new_sim.fields_inside[wl].efield
                h_in_1 = _new_sim.fields_inside[wl].hfield
                e_in_2 = other_sim.fields_inside[wl].efield
                h_in_2 = other_sim.fields_inside[wl].hfield
                e_in = torch.cat([e_in_1, e_in_2], dim=1)
                h_in = torch.cat([h_in_1, h_in_2], dim=1)

                pos1 = _new_sim.get_all_positions()
                pos2 = other_sim.get_all_positions()
                pos = torch.cat([pos1, pos2], dim=0)

                _f = Field(pos, e_in, h_in)
                fields_inside_combined[float(wl)] = _f

            _new_sim.fields_inside = fields_inside_combined

        # combine structures
        _new_sim.structures += other_sim.structures  # add 2 lists

        return _new_sim

    def copy(self, positions=None):
        """(batch) copy simulation

        optionally, a simulations with multiple copies of its structures can be created, where each copy of structures is shifted to a new position.

        Args:
            positions (list, optional): list of new positions to create copies of the structures at. If None, create a single, identical copy. Defaults to None.

        Returns:
            :class:`Simulation`: new simulation
        """
        if positions is None:
            import copy

            return copy.deepcopy(self)
        else:
            # generate multiple copies, moved to `positions`
            positions = torch.as_tensor(
                positions, device=self.device, dtype=DTYPE_FLOAT
            )

            # single position: expand dim
            if len(positions.shape) == 1:
                positions = positions.unsqueeze(0)

            assert len(positions.shape) == 2
            assert positions.shape[1] == 3

            new_sim_list = []
            for _r in positions:
                _sim = self.copy()
                for i in range(len(_sim.structures)):
                    _sim.structures[i] = _sim.structures[i] + _r
                new_sim_list.append(_sim)

            new_sim = new_sim_list.pop(0)
            for _s in new_sim_list:
                new_sim = new_sim.combine(_s)

            return new_sim

    def __add__(self, other):
        if issubclass(type(other), type(self)):
            # add another simulation: try combining both.
            return self.combine(other)
        else:
            raise ValueError(
                "Unknown add type, only addition of other sim is supported."
            )

    # --- convenience plotting wrapper
    def plot_structure(self, **kwargs):
        from torchgdm.visu import visu2d

        visu2d.structure(self, **kwargs)

    def plot_contour(self, **kwargs):
        from torchgdm.visu import visu2d

        visu2d.contour(self, **kwargs)

    def plot_structure_3d(self, **kwargs):
        from torchgdm.visu import visu3d

        visu3d.structure(self, **kwargs)

    def plot_efield_vectors_inside(self, wavelength, illumination_index=0, **kwargs):
        from torchgdm.visu import visu2d

        visu2d.vectorfield_inside(
            self,
            wavelength=wavelength,
            illumination_index=illumination_index,
            whichfield="e",
            **kwargs,
        )

    def plot_hfield_vectors_inside(self, wavelength, illumination_index=0, **kwargs):
        from torchgdm.visu import visu2d

        visu2d.vectorfield_inside(
            self,
            wavelength=wavelength,
            illumination_index=illumination_index,
            whichfield="h",
            **kwargs,
        )

    def plot_efield_vectors_inside_3d(
        self, wavelength, illumination_index=0, scale=1, **kwargs
    ):
        from torchgdm.visu import visu3d

        visu3d.vectorfield_inside(
            self,
            wavelength=wavelength,
            illumination_index=illumination_index,
            whichfield="e",
            scale=25 * scale,
            **kwargs,
        )

    def plot_hfield_vectors_inside_3d(
        self, wavelength, illumination_index=0, scale=1, **kwargs
    ):
        from torchgdm.visu import visu3d

        visu3d.vectorfield_inside(
            self,
            wavelength=wavelength,
            illumination_index=illumination_index,
            whichfield="h",
            scale=25 * scale,
            **kwargs,
        )

    # --- functions combining properties from all structures
    def get_all_positions(self) -> torch.Tensor:
        """return a list of all positions with dipoles"""
        return torch.cat([struct.positions for struct in self.structures], dim=0)

    def get_source_validity_radius(self) -> torch.Tensor:
        """return a list of all effective step-radii / enclosing sphere radius"""
        all_pos_step = [_s.get_source_validity_radius() for _s in self.structures]
        return torch.cat(all_pos_step)

    def get_closest_wavelength(self, wavelength):
        """return closest wavelength match available in the simulation"""
        from torchgdm.tools.misc import get_closest_wavelength

        return get_closest_wavelength(self, wavelength)

    # --- internal API: helper for polarizable / non-polarizable dipole handling
    def _get_polarizable_positions_indices_separate_p_m(
        self, structure_index=-1
    ) -> tuple:
        """get lists with indices of polarizable positions, separate for p and m

        Args:
            structure_index (int, optional): Optionally select a specific sub-structure. if -1, return for all structures. Defaults to -1.

        Returns:
            tuple: (torch.Tensor, torch.Tensor). position indices at which polarizable p and m dipoles are located
        """
        # optional: consider only a specific structure
        if structure_index == -1:
            structures = self.structures
        else:
            structures = [self.structures[structure_index]]

        idx_p = [torch.tensor([], dtype=torch.int32, device=self.device)]
        idx_m = [torch.tensor([], dtype=torch.int32, device=self.device)]
        i_offset = 0
        for struct in structures:
            N_pos = len(struct.positions)

            _sT = struct.evaluation_terms
            if "E" in _sT:
                idx_p.append(
                    torch.arange(i_offset, i_offset + N_pos, device=self.device)
                )
            if "H" in _sT:
                idx_m.append(
                    torch.arange(i_offset, i_offset + N_pos, device=self.device)
                )

            i_offset += N_pos

        return torch.cat(idx_p, dim=0), torch.cat(idx_m, dim=0)

    def _get_nonpolarizable_positions_indices_separate_p_m(
        self, structure_index=-1
    ) -> tuple:
        """get lists with indices of non-polarizable positions (where *no* polarizabilites are located), separate for p and m

        Args:
            structure_index (int, optional): Optionally select a specific sub-structure. if -1, return for all structures. Defaults to -1.

        Returns:
            tuple: (torch.Tensor, torch.Tensor). non-polarizable position indices at which *no* p or *no* m dipoles are located
        """
        # optional: consider only a specific structure
        if structure_index == -1:
            structures = self.structures
        else:
            structures = [self.structures[structure_index]]

        idx_p = [torch.tensor([], dtype=torch.int32, device=self.device)]
        idx_m = [torch.tensor([], dtype=torch.int32, device=self.device)]
        i_offset = 0
        for struct in structures:
            N_pos = len(struct.positions)

            _sT = struct.evaluation_terms
            if "E" not in _sT:
                idx_p.append(
                    torch.arange(i_offset, i_offset + N_pos, device=self.device)
                )
            if "H" not in _sT:
                idx_m.append(
                    torch.arange(i_offset, i_offset + N_pos, device=self.device)
                )

            i_offset += N_pos

        return torch.cat(idx_p, dim=0), torch.cat(idx_m, dim=0)

    def _get_polarizable_positions_separate_p_m(self) -> tuple:
        """get lists with polarizable positions, separate for p and m

        Args:
            structure_index (int, optional): Optionally select a specific sub-structure. if -1, return for all structures. Defaults to -1.

        Returns:
            tuple: (torch.Tensor, torch.Tensor). positions at which polarizable p and m dipoles are located
        """
        idx_p, idx_m = self._get_polarizable_positions_indices_separate_p_m()
        dp_pos = self.get_all_positions()
        pos_p = torch.index_select(dp_pos, 0, idx_p)
        pos_m = torch.index_select(dp_pos, 0, idx_m)

        return pos_p, pos_m

    def _get_nonpolarizable_positions_separate_p_m(self) -> tuple:
        """get lists with non-polarizable positions (where *no* polarizabilites are located), separate for p and m

        Args:
            structure_index (int, optional): Optionally select a specific sub-structure. if -1, return for all structures. Defaults to -1.

        Returns:
            tuple: (torch.Tensor, torch.Tensor). non-polarizable positions at which *no* p or *no* m dipoles are located
        """
        idx_p, idx_m = self._get_nonpolarizable_positions_indices_separate_p_m()
        dp_pos = self.get_all_positions()
        pos_miss_p = torch.index_select(dp_pos, 0, idx_p)
        pos_miss_m = torch.index_select(dp_pos, 0, idx_m)

        return pos_miss_p, pos_miss_m

    def _get_polarizable_mask_full_fields(self) -> torch.Tensor:
        # loop through structures and create field-mask for polarizable positions
        mask_fields = []
        for struct in self.structures:
            Npos = len(struct.positions)

            _sT = struct.evaluation_terms
            _mask_e0h0 = torch.zeros(
                (len(self.illumination_fields), Npos, 6),
                dtype=torch.bool,
                device=self.device,
            )
            if "E" in _sT:
                _mask_e0h0[..., :3] = 1
            if "H" in _sT:
                _mask_e0h0[..., 3:] = 1
            mask_fields.append(_mask_e0h0)
        mask_fields = torch.cat(mask_fields, dim=1)

        return mask_fields

    # --- polarizabilities of all structure dipoles
    def _get_all_polarizabilitites_6x6(self, wavelength: float) -> torch.Tensor:
        pola_list = torch.cat(
            [
                struct.get_polarizability_6x6(wavelength, self.environment)
                for struct in self.structures
            ],
            dim=0,
        )
        return pola_list

    def _get_all_polarizabilitites_pEH_3x6(self, wavelength: float) -> torch.Tensor:
        pola_list = torch.cat(
            [
                struct.get_polarizability_pEH_3x6(wavelength, self.environment)
                for struct in self.structures
            ],
            dim=0,
        )
        return pola_list

    def _get_all_polarizabilitites_mEH_3x6(self, wavelength: float) -> torch.Tensor:
        pola_list = torch.cat(
            [
                struct.get_polarizability_mEH_3x6(wavelength, self.environment)
                for struct in self.structures
            ],
            dim=0,
        )
        return pola_list

    def _get_all_selfterms_6x6(self, wavelength: float) -> torch.Tensor:
        st_list = []
        for struct in self.structures:
            st_list.append(struct.get_selfterm_6x6(wavelength, self.environment))
        return torch.cat(st_list, dim=0)

    # --- illumination fields at all dipoles
    def _get_all_e0(self, wavelength: float) -> torch.Tensor:
        dp_pos = self.get_all_positions()
        e0 = get_e0(sim=self, wavelength=wavelength, r_probe=dp_pos)
        return e0

    def _get_all_h0(self, wavelength: float) -> torch.Tensor:
        dp_pos = self.get_all_positions()
        h0 = get_h0(sim=self, wavelength=wavelength, r_probe=dp_pos)
        return h0

    def _get_all_e0_h0(self, wavelength: float) -> torch.Tensor:
        e0 = self._get_all_e0(wavelength)
        h0 = self._get_all_h0(wavelength)
        e0h0 = torch.cat([e0, h0], dim=-1)
        return e0h0

    # --- fields per structure
    def _get_insidefields_for_single_structure(self, structure_index):
        # get indices of positions corresponding to selected structure
        n_pos = [len(_s.positions) for _s in self.structures]
        n_pos_cum = torch.cumsum(torch.as_tensor([0] + n_pos), 0)
        idx = [
            torch.arange(_nstart, _nstart + _n) for _nstart, _n in zip(n_pos_cum, n_pos)
        ][structure_index]
        idx = idx.to(device=self.device)

        # extract inside fields of selected structure
        fields_inside_struct = dict()
        for wl in self.wavelengths:
            wl = float(wl)

            e_in_mask = torch.index_select(self.fields_inside[wl].efield, 1, idx)
            h_in_mask = torch.index_select(self.fields_inside[wl].hfield, 1, idx)

            _f = Field(self.structures[structure_index].positions, e_in_mask, h_in_mask)
            fields_inside_struct[float(wl)] = _f

        return fields_inside_struct

    # --- fields and dipole moments at polarizable sources only
    def _get_polarizablefields_e0_h0(self, wavelength: float) -> torch.Tensor:
        # get full illumination E0, H0 fields (defined in parent class)
        e0_h0_all = self._get_all_e0_h0(wavelength=wavelength)

        mask_fields = self._get_polarizable_mask_full_fields()
        return e0_h0_all[mask_fields].reshape(len(self.illumination_fields), -1)

    def _get_polarizablefields_e(self, wavelength):
        wl_key = float(wavelength)
        idx_p, idx_m = self._get_polarizable_positions_indices_separate_p_m()
        e_in_mask = torch.index_select(self.fields_inside[wl_key].efield, 1, idx_p)
        return e_in_mask

    def _get_polarizablefields_h(self, wavelength):
        wl_key = float(wavelength)
        idx_p, idx_m = self._get_polarizable_positions_indices_separate_p_m()
        h_in_mask = torch.index_select(self.fields_inside[wl_key].hfield, 1, idx_m)
        return h_in_mask

    def _get_polarizable_dipole_moments_p(self, wavelength):
        wl_key = float(wavelength)
        idx_p, idx_m = self._get_polarizable_positions_indices_separate_p_m()

        e_in_mask = self._get_polarizablefields_e(wavelength)
        e_in_mask = torch.index_select(self.fields_inside[wl_key].efield, 1, idx_p)
        h_in_mask = torch.index_select(self.fields_inside[wl_key].hfield, 1, idx_p)
        e_h_in_mask = torch.cat([e_in_mask, h_in_mask], dim=-1)

        # calc. electric dipole moments due to E and H
        alpha_pEH = self._get_all_polarizabilitites_pEH_3x6(wavelength)
        alpha_pEH_mask = torch.index_select(alpha_pEH, 0, idx_p)
        p_in_mask = torch.matmul(alpha_pEH_mask, e_h_in_mask.unsqueeze(-1))

        return p_in_mask[..., 0]  # remove last dim from broadcast in matmul

    def _get_polarizable_dipole_moments_m(self, wavelength):
        wl_key = float(wavelength)
        idx_p, idx_m = self._get_polarizable_positions_indices_separate_p_m()

        e_in_mask = torch.index_select(self.fields_inside[wl_key].efield, 1, idx_m)
        h_in_mask = torch.index_select(self.fields_inside[wl_key].hfield, 1, idx_m)
        e_h_in_mask = torch.cat([e_in_mask, h_in_mask], dim=-1)

        # calc. magnetic dipole moments due to E and H
        alpha_mEH = self._get_all_polarizabilitites_mEH_3x6(wavelength)
        alpha_mEH_mask = torch.index_select(alpha_mEH, 0, idx_m)
        m_in_mask = torch.matmul(alpha_mEH_mask, e_h_in_mask.unsqueeze(-1))

        return m_in_mask[..., 0]  # remove last dim from broadcast in matmul


# --- simulation container class
class Simulation(SimulationBase):
    """torchgdm simulation container"""

    __name__ = "torchgdm simulation container"

    def __init__(
        self,
        structures: list,
        environment,
        illumination_fields: list,
        wavelengths: list,
        linearsystem="default",
        test_structure_collisions: bool = True,
        on_distance_violation: str = "error",
        device: torch.device = None,
    ):
        """simulation constructor

        Args:
            structures (list): list of structures
            environment (environmentBase): environment class
            illumination_fields (list): list of illuminations
            wavelengths (list): list of wavelengths to evaluate (nm)
            linearsystem (linearSystemBase): solver for the coupling system. Defaults to `LinearSystemFullMoreMemoryEfficient`
            test_structure_collisions (bool, optional): Check for overlapping / colliding structures. Can be slow in case of many structures. Defaults to True.
            on_distance_violation (str, optional): behavior on distance violation. can be "error", "warn" or None (do nothing). Defaults to "error".
            device (torch.device, optional): Defaults to 'cpu'.
        """
        super().__init__(
            structures=structures,
            environment=environment,
            illumination_fields=illumination_fields,
            wavelengths=wavelengths,
            device=device,
            test_structure_collisions=test_structure_collisions,
            on_distance_violation=on_distance_violation,
        )
        if type(linearsystem) == str:
            if linearsystem.lower() == "default":
                from torchgdm.linearsystem import LinearSystemFullMoreMemoryEfficient

                self.linearsystem = LinearSystemFullMoreMemoryEfficient(device=device)
            else:
                raise ValueError("Unknown linearsystem.")
        else:
            self.linearsystem = linearsystem

        # set the global device for the simulation
        self.set_device(self.device)

    def __repr__(self):
        out_str = ""
        out_str += "simulation on device '{}'...\n".format(self.device)

        out_str += " - {}\n".format(self.environment.__name__)

        out_str += " - {} structures\n".format(len(self.structures))

        _pos = self.get_all_positions()
        out_str += " - {} positions\n".format(len(_pos))

        _pos_p, _pos_m = self._get_polarizable_positions_separate_p_m()
        out_str += " - {} coupled dipoles\n".format(len(_pos_p) + len(_pos_m))
        out_str += "   (of which {} p, {} m)\n".format(len(_pos_p), len(_pos_m))

        out_str += " - {} wavelengths\n".format(len(self.wavelengths))

        out_str += " - {} illumination fields per wavelength".format(
            len(self.illumination_fields)
        )
        return out_str

    def set_device(self, device):
        super().set_device(device)
        self.linearsystem.set_device(device)

    def reset_fields(self):
        self.fields_inside = dict()

    def run(
        self,
        batch_size: int = 32,
        calc_missing: bool = False,
        verbose: int = 1,
        progress_bar=True,
    ):
        """run the simulation: evaluate all illuminations at all wavelengths

        results are the selfconsistent, full electric and magnetic fields
        at each meshpoint of all structures.

        this will populate `self.fields_inside` a dict with the wavelengths (in nm) as keys.
        Each element is a :class:`torchgdm.Field` instance, containing the `efield` and `hfield`.

        Args:
            batch_size (int, optional):    Nr of incident fields evaluated in parallel. Defaults to 32.
            calc_missing (bool, optional): calculate non-relevant internal fields, typically
                                           H-Field in discretized structures). Defaults to False.
            verbose (int, optional):       print status info. Defaults to 1.
        """
        t_start = time.time()
        if verbose:
            print("-" * 60)
            print(self)
            print("\nrun spectral simulation:")

        # loop over wavelengths
        for wl in tqdm(self.wavelengths, progress_bar, title=""):
            t_start_wl = time.time()
            if verbose and not progress_bar:
                print("wl={:.1f}nm: ".format(wl), end="")
            e_in, h_in = self.linearsystem.solve(
                sim=self,
                wavelength=wl,
                batch_size=batch_size,
                verbose=verbose,
            )

            _f = Field(self.get_all_positions(), e_in, h_in)
            self.fields_inside[float(wl)] = _f

            # pure electric / pure magnetic reponse: calc. missing fields via propagation
            if calc_missing:
                t_fillmiss0 = time.time()
                if verbose and not progress_bar:
                    print(" propa", end="")
                self._calculate_fields_at_nonpolarizable(wavelength=wl)
                t_fillmiss1 = time.time()
                if verbose and not progress_bar:
                    print(" {:.2f}s.".format(t_fillmiss1 - t_fillmiss0), end="")

            # cleanup
            purge_mem(dev=e_in.device)

            if verbose and not progress_bar:
                t_end = time.time()
                print(" tot {:.2f}s.".format(t_end - t_start_wl))

        if verbose:
            t_end = time.time()
            if not progress_bar:
                print()
            print("spectrum done in {:.2f}s".format(t_end - t_start))
            print("-" * 60 + "\n")

    # --- fields propagation
    def _calculate_fields_at_nonpolarizable(self, wavelength: float):
        """calculate internal fields for non-polarizable locations

        calculates the magnetic field at locations of purely electric-electric polarizabilties and
        the electric field at locations of purely magnetic-magnetic polarizabilties

        Args:
            wavelength (float): in nm
        """
        # - calculate missing fields at a wavelength via re-propagation
        from torchgdm.postproc.fields import _nearfield_e, _nearfield_h

        wl_key = float(wavelength)
        field_shape = self.fields_inside[wl_key].efield.shape

        mask = self._get_polarizable_mask_full_fields()
        mask_e, mask_h = torch.chunk(mask, 2, dim=2)

        (
            pos_e_miss,
            pos_h_miss,
        ) = self._get_nonpolarizable_positions_separate_p_m()

        if len(pos_e_miss) > 0:
            nf_e = _nearfield_e(
                sim=self,
                wavelength=wavelength,
                r_probe=pos_e_miss,
                source_distance_steps=0.1,  # exclude 'self-source' (divergence of G)
            )
            _e = nf_e["e_tot"]
            e_in = self.fields_inside[wl_key].efield.flatten()
            e_in[~mask_e.flatten()] = _e.flatten()
            e_in = e_in.reshape(field_shape)
            self.fields_inside[wl_key].set_efield(e_in)

        if len(pos_h_miss) > 0:
            nf_h = _nearfield_h(
                sim=self,
                wavelength=wavelength,
                r_probe=pos_h_miss,
                source_distance_steps=0.1,  # exclude 'self-source' (divergence of G)
            )
            _h = nf_h["h_tot"]
            h_in = self.fields_inside[wl_key].hfield.flatten()
            h_in[~mask_h.flatten()] = _h.flatten()
            h_in = h_in.reshape(field_shape)
            self.fields_inside[wl_key].set_hfield(h_in)

    # --- wrapper for post processing functions
    def get_geometric_crosssection(self, projection="xy"):
        """get ensemble geometric cross section of all structures in nm^2

        Args:
            projection (str, optional): cartesian projection of cross section. Defaults to "xy"

        Returns:
            float: geometric cross section in nm^2
        """
        import torchgdm as tg

        pos = self.get_all_positions()
        steps = torch.cat(
            [s.step / s.mesh_normalization_factor for s in self.structures]
        )
        return tg.tools.geometry._get_geo_cs_positions_steps(pos, steps, projection)

    # - fields
    def get_nearfield(
        self, wavelength, r_probe=None, illumination_index=None, **kwargs
    ):
        """calculate the nearfield (electric and magnetic) at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident fields (keys: "tot", "sca", "inc") in instances of :class:`torchgdm.Field`
        """
        import torchgdm as tg

        if r_probe is None:
            r_probe = self.get_all_positions()

        nf = tg.postproc.fields.nf(
            self,
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        return nf

    def get_nf(self, **kwargs):
        """alias for :meth:`Simulation.get_nearfield`"""
        return self.get_nearfield(**kwargs)

    def get_field_gradients(
        self,
        wavelength: float,
        r_probe: torch.Tensor = None,
        illumination_index: int = None,  # None: batch all illumination fields
        source_distance_steps: float = None,
        delta=0.1,
        whichfield="e_tot",
        **kwargs,
    ):
        """nearfield gradients inside or in proximity of a nanostructure

        Calculate field-gradients (positions defined by `r_probe`).
        pytorch AD is not efficient for gradients of functions R^n --> R^m, with $n \\sim m >> 1$.
        Therefore, here numerical derivatives are calculated via center differences.

        Based on the original implementation in pyGDM by C. Majorel.

        Warning: The current implementation is not memory efficient, since all fields are calculated, even though not required.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): in nm
            r_probe (torch.Tensor, optional): tuple (x,y,z) or list of 3-lists/-tuples to evaluate field gradients. Use all structure positions by default. Defaults to None.
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            delta (float, optional):  differential step for numerical center-derivative (in nanometers). Defaults to 0.1.
            whichfield (str, optional): fields to calculate the gradient for. One of ["e_sca","e_tot","e_inc", "h_sca","h_tot,"h_inc], . Defaults to "e_tot".

        Raises:
            ValueError: _description_

        Returns:
            3 lists of 3-tuples [dAx, dAy, dAz] (complex): dAj are the differential terms:
                - idx [0] = dE/dx = [dEx/dx, dEy/dx, dEz/dx]
                - idx [1] = dE/dy = [dEx/dy, dEy/dy, dEz/dy]
                - idx [2] = dE/dz = [dEx/dz, dEy/dz, dEz/dz]
        """
        import torchgdm as tg

        if r_probe is None:
            r_probe = self.get_all_positions()
        # field_gradient(sim, wavelength, r_probe, delta=delta, whichfield=field)
        fieldgrad = tg.postproc.fields.field_gradient(
            self,
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            source_distance_steps=source_distance_steps,
            delta=delta,
            whichfield=whichfield,
            **kwargs,
        )

        return fieldgrad

    def get_nearfield_intensity_efield(
        self, wavelength, r_probe=None, illumination_index=None, **kwargs
    ):
        """calculate the electric field intensity in the nearfield at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident electric field intensity (keys: "tot", "sca", "inc")
        """
        nf = self.get_nearfield(
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        # add positions
        nf["r_probe"] = nf["tot"].get_positions()

        for which in ["tot", "sca", "inc"]:
            nf[which] = nf[which].get_efield_intensity()

        return nf

    def get_nearfield_intensity_hfield(
        self, wavelength, r_probe=None, illumination_index=None, **kwargs
    ):
        """calculate the magnetic field intensity in the nearfield at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident magnetic field intensity (keys: "tot", "sca", "inc")
        """
        nf = self.get_nearfield(
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        # add positions
        nf["r_probe"] = nf["tot"].get_positions()

        for which in ["tot", "sca", "inc"]:
            nf[which] = nf[which].get_hfield_intensity()

        return nf

    def get_chirality(
        self, wavelength, r_probe=None, illumination_index=None, **kwargs
    ):
        """calculate the nearfield chirality at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident magnetic field intensity (keys: "tot", "sca", "inc")
        """
        nf_C = self.get_nearfield(
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        # add positions
        nf_C["r_probe"] = nf_C["tot"].get_positions()

        # overwrite Field instances by the chirality
        for which in ["tot", "sca", "inc"]:
            nf_C[which] = nf_C[which].get_chirality()

        return nf_C

    def get_poynting(self, wavelength, r_probe=None, illumination_index=None, **kwargs):
        """calculate the Poynting vector at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident field's Poynting vector (keys: "tot", "sca", "inc")
        """
        nf_S = self.get_nearfield(
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        # add positions
        nf_S["r_probe"] = nf_S["tot"].get_positions()

        # overwrite Field instances by the Poynting vectors
        for which in ["tot", "sca", "inc"]:
            nf_S[which] = nf_S[which].get_poynting()

        return nf_S

    def get_energy_flux(
        self, wavelength, r_probe=None, illumination_index=None, **kwargs
    ):
        """calculate the energy flux (time averaged Poynting vector) at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident field's time averaged Poynting vector (keys: "tot", "sca", "inc")
        """
        nf_S = self.get_nearfield(
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        # add positions
        nf_S["r_probe"] = nf_S["tot"].get_positions()

        # overwrite Field instances by the time averaged Poynting vectors
        for which in ["tot", "sca", "inc"]:
            nf_S[which] = nf_S[which].get_energy_flux()

        return nf_S

    def get_green(
        self,
        r_probe: torch.Tensor,
        r_source: torch.Tensor,
        wavelength: float,
        use_sim_copy=True,
        progress_bar=True,
        verbose=0,
        **kwargs,
    ):
        """calculate the Green's tensor in presence of the structure(s)

        kwargs (such as `batch_size`) are passed to :func:`tg.postproc.green.G`.

        Args:
            sim (`torchgdm.Simulation`): simulation
            r_probe (torch.Tensor): probe position(s)
            r_source (torch.Tensor): source location(s)
            wavelength (float): in nm
            use_sim_copy (bool, optional): Use copy of simulation or use simulation in-place. May not work in some autograd scenario. Defaults to True.
            progress_bar (bool, optional): Show progress bar. Defaults to True.
            verbose (bool, optional): Print status of underlying simulation call. Defaults to False.

        Returns:
            dict: Green's tensors for all source/probe position combinations:
                - "G_6x6": full 6x6 tensor
                - "G_Ep": electric-electric 3x3 tensor
                - "G_Em": electric-magnetic 3x3 tensor
                - "G_Hp": magnetic-electric 3x3 tensor
                - "G_Hm": magnetic-magnetic 3x3 tensor
        """
        import torchgdm as tg

        if r_probe is None:
            r_probe = self.get_all_positions()

        G = tg.postproc.green.G(
            self,
            r_probe=r_probe,
            r_source=r_source,
            wavelength=wavelength,
            use_sim_copy=use_sim_copy,
            progress_bar=progress_bar,
            verbose=verbose,
            **kwargs,
        )

        return G

    def get_ldos(
        self,
        r_probe: torch.Tensor,
        wavelength: float,
        progress_bar=True,
        verbose=False,
        **kwargs,
    ):
        """calculate the LDOS close to the structure(s)

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (`torchgdm.Simulation`): simulation
            r_probe (torch.Tensor): probe position(s)
            wavelength (float): in nm
            progress_bar (bool, optional): Show progress bar. Defaults to True.
            verbose (bool, optional): Print status info. Defaults to False.


        Returns:
            dict: Green's tensors, partial and averaged LDOS all probe positions:
                - "G_ii": full 6x6 tensor
                - "ldos_partial": partial electric (first 3) and magnetic (last 3) LDOS. (Diagonal elements of Green's tensor)
                - "ldos_e": full electric LDOS
                - "ldos_m": full magnetic LDOS
        """
        import torchgdm as tg

        if r_probe is None:
            r_probe = self.get_all_positions()

        ldos = tg.postproc.green.ldos(
            self,
            r_probe=r_probe,
            wavelength=wavelength,
            progress_bar=progress_bar,
            verbose=verbose,
            **kwargs,
        )

        return ldos

    def get_farfield(self, wavelength, r_probe, illumination_index=None, **kwargs):
        """calculate the far-field (electric and magnetic) at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            wavelength (float): evaluation wavelength (in nm)
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case full hemisphere coordinates are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: results for total, scattered and incident fields in instances of :class:`torchgdm.Field`
        """
        import torchgdm as tg

        ff = tg.postproc.fields.ff(
            self,
            wavelength=wavelength,
            r_probe=r_probe,
            illumination_index=illumination_index,
            **kwargs,
        )

        return ff

    # - rasterscan
    def get_rasterscan(self, func, wavelength, kw_r="r_focus", **kwargs):
        """Get rasterscan from a "scan-position" parameters `kw_r` and evaluation function `func`

        `kw_r` must be an attribute of the available illuminations, defining
        some position (3-tuple), this is used as the raster-scan position for each evaluation.

        kwargs are passed to `func`.

        Args:
            func (callable): postprocessing function, from :mod:`tg.postproc`
            wavelength (float): evaluation wavelength (in nm)
            kw_r (str): kwarg used for scan position of rasterscan

        Returns:
            dict: spectrum results
        """
        from torchgdm.tools.batch import calc_rasterscan

        rs_res = calc_rasterscan(self, func, wavelength=wavelength, kw_r=kw_r)
        return rs_res

    # - spectra
    def get_spectra(self, func, **kwargs):
        """Get spectra using all available wavelengths for evaluation function `func`

        kwargs are passed to `func`

        Args:
            func (callable): postprocessing function, from :mod:`tg.postproc`

        Returns:
            dict: spectrum results
        """
        from torchgdm.tools.batch import calc_spectrum

        nf_spec = calc_spectrum(self, func, **kwargs)
        return nf_spec

    def get_spectra_scs(self, **kwargs):
        """total scattering, absorption and extinction cross section spectra

        This function also returns extinction and absorption cross-sections, which are calculated as byproducts.

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.crosssect.scs`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.
            radiative_correction (bool, optional): Whether to add radiative correction term. Defaults to True.

        Returns:
            dict: dictionaty containing the scattering, absorption and extinction cross section spectra
        """
        import torchgdm as tg

        scs_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.crosssect.scs, **kwargs
        )
        return scs_spec

    def get_crosssections(self, **kwargs):
        """total scattering, absorption and extinction cross section spectra

        *Note:* this is an alias for :func:`Simulation.get_spectra_scs`.

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.crosssect.scs`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.
            radiative_correction (bool, optional): Whether to add radiative correction term. Defaults to True.

        Returns:
            dict: dictionaty containing the scattering, absorption and extinction cross section spectra
        """
        return self.get_spectra_scs(**kwargs)

    def get_spectra_ecs(self, **kwargs):
        """total exctinction cross section spectra

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.crosssect.ecs`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.

        Returns:
            dict: dictionaty containing the extinction cross section spectra
        """
        import torchgdm as tg

        ecs_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.crosssect.ecs, **kwargs
        )
        return ecs_spec

    def get_spectra_acs(self, **kwargs):
        """total absorption cross section spectra

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.crosssect.acs`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            normalize_to_local_e0 (bool, optional): Whether to normalize to local illumination amplitude. Can be useful to get an approximate cross-section for non-plane wave illuminations. Defaults to False.

        Returns:
            dict: dictionaty containing the absorption cross section spectra
        """
        import torchgdm as tg

        ecs_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.crosssect.ecs, **kwargs
        )
        return ecs_spec

    def get_spectra_nf(self, r_probe=None, **kwargs):
        """Get spectra of complex nearfields at all available wavelengths at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: contains results for total, scattered and incident fields in instances of :class:`torchgdm.Field`
        """
        import torchgdm as tg

        if r_probe is None:
            r_probe = self.get_all_positions()

        nf_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.fields.nf, r_probe=r_probe, **kwargs
        )
        return nf_spec

    def get_spectra_nf_intensity_e(self, r_probe=None, **kwargs):
        """Get spectrum of integrated electric nearfield intensities

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: contains results for total (key "tot"), scattered (key "sca") and incident (key "inc") electric field intensity
        """
        nf_spec = self.get_spectra_nf(r_probe=r_probe, **kwargs)

        for which in ["tot", "sca", "inc"]:
            nf_spec[which] = torch.stack(
                [f.get_integrated_efield_intensity() for f in nf_spec[which]]
            )

        return nf_spec

    def get_spectra_nf_intensity_h(self, r_probe=None, **kwargs):
        """Get spectrum of integrated magnetic nearfield intensities

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.nf`.

        Args:
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: contains results for total (key "tot"), scattered (key "sca") and incident (key "inc") magnetic field intensity
        """
        nf_spec = self.get_spectra_nf(r_probe=r_probe, **kwargs)

        for which in ["tot", "sca", "inc"]:
            nf_spec[which] = torch.stack(
                [f.get_integrated_hfield_intensity() for f in nf_spec[which]]
            )

        return nf_spec

    def get_spectra_nf_intensity_e_inside(self):
        """Get spectrum of integrated electric nearfield intensities inside structure"""
        import torchgdm as tg

        nf_int_e = tg.tools.batch.calc_spectrum(
            self, tg.postproc.fields.integrated_nf_intensity_e_inside
        )

        return nf_int_e

    def get_spectra_nf_intensity_h_inside(self):
        """Get spectrum of integrated magnetic nearfield intensities inside structure"""
        import torchgdm as tg

        nf_int_h = tg.tools.batch.calc_spectrum(
            self, tg.postproc.fields.integrated_nf_intensity_h_inside
        )

        return nf_int_h

    def get_spectra_ff(self, r_probe=None, **kwargs):
        """Get spectra of complex far-fields at all available wavelengths at positions `r_probe`

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.ff`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case all positions of the simulation structures are used
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: contains results for total, scattered and incident fields in instances of :class:`torchgdm.Field`
        """
        import torchgdm as tg

        ff_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.fields.ff, r_probe=r_probe, **kwargs
        )
        return ff_spec

    def get_spectra_ff_intensity(self, r_probe=None, **kwargs):
        """Get spectrum of integrated far-field intensities

        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.fields.ff`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            r_probe (dict or list, optional): probe locations (gen. from `tools.geometry`). Defaults to None, in which case intergration goes over full hemisphere
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

        Returns:
            dict: contains results for total (key "tot"), scattered (key "sca") and incident (key "inc") far-field intensity
        """
        ff_spec = self.get_spectra_ff(r_probe, **kwargs)

        for which in ["tot", "sca", "inc"]:
            ff_spec[which] = torch.stack(
                [f.get_integrated_efield_intensity() for f in ff_spec[which]]
            )

        return ff_spec

    def get_spectra_multipole(self, **kwargs):
        """Get spectrum of exact multipole decomposition

        This returns the spectra for the full multipole moments of dipole and quadrupole order.
        For more detailed documentation, see :func:`torchgdm.postproc.multipole.decomposition_exact`.
        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.multipole.decomposition_exact`.

        Args:
            sim (:class:`torchgdm.Simulation`): simulation instance
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            r0 (torch.Tensor, optional): [x,y,z] position of multipole decomposition development. Defaults to None, in which case the center of gravity is used.
            epsilon (float, optional): positions too close to r0 will moved away from r0 by epsilon (in units of step) to avoid numerical divergence of the Bessel terms. Defaults to 0.1.
            long_wavelength_approx (bool, optional): if True, use long wavelength approximation. Defaults to False.

        Returns:
            dict: each element contains the spectrum for one multipole moment. dipoles are rank-1 (3-vectors), quadrupoles are rank 2 (3x3 tensors):
                - 'ed_tot': electric dipole (full)
                - 'md': magnetic dipole
                - 'eq_tot': electric quadrupole (full)
                - 'mq': magnetic quadrupole
                - 'ed_1': electric dipole (first order)
                - 'ed_toroidal': toroidal dipole
                - 'eq1': electric quadrupole (first order)
                - 'eq_toroidal': toroidal quadrupole
        """
        import torchgdm as tg

        mpd_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.multipole.decomposition_exact, **kwargs
        )
        return mpd_spec

    def get_spectra_multipole_scs(self, **kwargs):
        """spectra of multipole decomposition of extinction cross section

        Returns the exact multipole decomposition of the extinction cross section spectra: electric and magnetic dipole and quadrupole moments.
        For more detailed documentation, see :func:`torchgdm.postproc.multipole.ecs`.
        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.multipole.ecs`.

        For details about the extinction section of multipole moments, see:
            Evlyukhin, A. B. et al. *Multipole analysis of light scattering by
            arbitrary-shaped nanoparticles on a plane surface.*,
            JOSA B 30, 2589 (2013)

        Args:
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            r0 (torch.Tensor, optional): [x,y,z] position of multipole decomposition development. Defaults to None, in which case the center of gravity is used.
            epsilon (float, optional): positions too close to r0 will moved away from r0 by epsilon (in units of step) to avoid numerical divergence of the Bessel terms. Defaults to 0.1.
            eps_dd (float, optional): numerical integration step for field gradients calc. (in nm). Required for e/m quadrupoles. Defaults to 0.1.
            normalization_E0 (bool, optional): Normalize to illumination amplitude at `r0`. Can be useful to get approximate results for non-plane wave illumination. Defaults to False.
            long_wavelength_approx (bool, optional): if True, use long wavelength approximation. Defaults to False.

        Returns:
            dict:
            - 'ecs_ed': electric dipole extinction cross section spectra (in nm^2)
            - 'ecs_md': magnetic dipole extinction cross section spectra (in nm^2)
            - 'ecs_eq': electric quadrupole extinction cross section spectra (in nm^2)
            - 'ecs_mq': magnetic quadrupole extinction cross section spectra (in nm^2)
        """
        import torchgdm as tg

        scs_mpd_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.multipole.scs, **kwargs
        )
        return scs_mpd_spec

    def get_spectra_multipole_ecs(self, **kwargs):
        """spectrum of multipole decomposition of scattering cross section

        Returns the exact multipole decomposition of the scattering cross section spectra: electric and magnetic dipole and quadrupole moments.
        For more detailed documentation, see :func:`torchgdm.postproc.multipole.scs`.
        kwargs (such as `batch_size`) are passed to :func:`torchgdm.postproc.multipole.scs`.

        For details about the exact multipole formalism and scs calculation, see:
            Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic
            multipole expansion beyond the long-wavelength approximation.*
            Optics Communications 407, 17–21 (2018)

        Args:
            illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
            with_toroidal (bool, optional): whether to add toroidal moments to electric dipole and quadrupole. Defaults to True.
            r0 (torch.Tensor, optional): [x,y,z] position of multipole decomposition development. Defaults to None, in which case the center of gravity is used.
            epsilon (float, optional): positions too close to r0 will moved away from r0 by epsilon (in units of step) to avoid numerical divergence of the Bessel terms. Defaults to 0.1.
            normalization_E0 (bool, optional): Normalize to illumination amplitude at `r0`. Can be useful to get approximate results for non-plane wave illumination. Defaults to False.
            long_wavelength_approx (bool, optional): if True, use long wavelength approximation. Defaults to False.

        Returns:
            dict:
            - 'scs_ed': electric dipole scattering cross section spectra (in nm^2)
            - 'scs_md': magnetic dipole scattering cross section spectra (in nm^2)
            - 'scs_eq': electric quadrupole scattering cross section spectra (in nm^2)
            - 'scs_mq': magnetic quadrupole scattering cross section spectra (in nm^2)
        """
        import torchgdm as tg

        scs_mpd_spec = tg.tools.batch.calc_spectrum(
            self, tg.postproc.multipole.ecs, **kwargs
        )
        return scs_mpd_spec



# -*- coding: utf-8 -*-
"""
numba implementations as well as vectorized numpy versions of 
free space Green's dyads for dipole and quadrupole sources
@authors: C. Majorel, P. R Wiecha
"""
#%%
import time
import math
import cmath

import numpy as np


def test_positional_input(R1: np.array, R2: np.array):
    if R1.shape != R2.shape:
        raise ValueError("R1 and R2 need same shape")
    if len(R1.shape) == 1:
        if R1.shape[0] != 3:
            raise ValueError(
                "single point needs to be cartesian coordinate with 3 values."
            )
        # expand dim
        R1 = R1[None, :]
        R2 = R2[None, :]
    elif len(R1.shape) == 2:
        if R1.shape[1] != 3:
            raise ValueError("points need to be cartesian coordinates with 3 values.")
    elif len(R1.shape) > 2:
        raise ValueError(
            "single point or list of points required, but got higher dimensional input data."
        )

    return R1, R2


## --- free space propagator dipole - numpy-vec
def G0_Ep_vec(
    R1: np.array,
    R2: np.array,
    wavelength: float,
    eps_env: float,
    dp_radius: float = 1e-20,
):
    R1, R2 = test_positional_input(R1, R2)

    R = R2 - R1
    I = np.identity(3)[None, :]  # , dtype='c16')

    lR2 = np.sum(R**2, axis=-1)
    lR = np.sqrt(lR2)
    lR[
        lR < dp_radius
    ] = 1.0  # add 1 to diag to avoid div by zero (treat self-terms separately later)

    k0 = 2 * math.pi / wavelength
    k = k0 * np.sqrt(eps_env)

    RR = np.einsum("...i,...j", R, R)

    R2diag = I * lR2[..., None, None]
    T1 = np.divide((RR - R2diag), np.power(lR, 3)[..., None, None])
    T2 = np.divide((3 * RR - R2diag), np.power(lR, 4)[..., None, None])
    T3 = np.divide((3 * RR - R2diag), np.power(lR, 5)[..., None, None])

    # phase term
    sG0 = np.exp(1j * k * lR) / eps_env

    G0 = (-np.power(k, 2) * T1 - 1j * k * T2 + T3) * sG0[..., None, None]

    return G0


def G0_Hp_vec(
    R1: np.array,
    R2: np.array,
    wavelength: float,
    eps_env: float,
    dp_radius: float = 1e-20,
):
    R1, R2 = test_positional_input(R1, R2)

    R = R2 - R1
    lR2 = np.sum(R**2, axis=-1)
    lR = np.sqrt(lR2)
    lR[
        lR < dp_radius
    ] = 1.0  # add 1 to diag to avoid div by zero (treat self-terms separately later)

    k0 = 2 * np.pi / wavelength
    k02n = np.sqrt(eps_env) * k0**2

    Dz = R[..., 2]
    Dy = R[..., 1]
    Dx = R[..., 0]

    T2XY = Dz / lR2
    T3XY = Dz / (lR**3)
    T2XZ = -Dy / lR2
    T3XZ = -Dy / (lR**3)
    T2YZ = Dx / lR2
    T3YZ = Dx / (lR**3)

    # Initialize and fill tensor
    G0 = np.zeros((*R.shape[:-1], 3, 3), dtype=np.complex_)
    G0[..., 0, 1] = 1j * k0 * T3XY + k02n * T2XY
    G0[..., 0, 2] = 1j * k0 * T3XZ + k02n * T2XZ
    G0[..., 1, 2] = 1j * k0 * T3YZ + k02n * T2YZ
    G0[..., 1, 0] = -G0[..., 0, 1]
    G0[..., 2, 0] = -G0[..., 0, 2]
    G0[..., 2, 1] = -G0[..., 1, 2]

    # phase term
    cG0 = -1 * np.exp(1j * k0 * np.sqrt(eps_env) * lR)
    G0 *= cG0[..., None, None]

    return G0


def G0_Hm_vec(R1, R2, wavelength, eps_env, dp_radius: float = 1e-20):
    return G0_Ep_vec(R1, R2, wavelength, eps_env, dp_radius)


def G0_Em_vec(R1, R2, wavelength, eps_env, dp_radius: float = 1e-20):
    return -1 * G0_Hp_vec(R1, R2, wavelength, eps_env, dp_radius)


## --- free space propagator dipole
def G0_Ep(R1, R2, wavelength, eps):
    """
    R1: dipole position
    R2: evaluation position
    """
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] - R1[2]
    lR = math.sqrt(Dx**2 + Dy**2 + Dz**2)

    k = 2 * np.pi / wavelength
    cn = cmath.sqrt(eps)
    ck0 = 1j * k * cn
    k2 = k * k * eps

    r25 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 2.5)
    r2 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 2.0)
    r15 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 1.5)

    #!C-------------------------------------------------------------------
    T1XX = -1 * (Dy * Dy + Dz * Dz) / r15
    T2XX = (2 * Dx * Dx - Dy * Dy - Dz * Dz) / r2
    T3XX = (2 * Dx * Dx - Dy * Dy - Dz * Dz) / r25
    #!C-------------------------------------------------------------------
    T1XY = Dx * Dy / r15
    T2XY = 3 * Dx * Dy / r2
    T3XY = 3 * Dx * Dy / r25
    #!C-------------------------------------------------------------------
    T1XZ = Dx * Dz / r15
    T2XZ = 3 * Dx * Dz / r2
    T3XZ = 3 * Dx * Dz / r25
    #!C-------------------------------------------------------------------
    T1YY = -(Dx * Dx + Dz * Dz) / r15
    T2YY = (2 * Dy * Dy - Dx * Dx - Dz * Dz) / r2
    T3YY = (2 * Dy * Dy - Dx * Dx - Dz * Dz) / r25
    #!C-------------------------------------------------------------------
    T1YZ = Dy * Dz / r15
    T2YZ = 3 * Dy * Dz / r2
    T3YZ = 3 * Dy * Dz / r25
    #!C------------------------------------------------------------------
    T1ZZ = -(Dx * Dx + Dy * Dy) / r15
    T2ZZ = (2 * Dz * Dz - Dx * Dx - Dy * Dy) / r2
    T3ZZ = (2 * Dz * Dz - Dx * Dx - Dy * Dy) / r25

    CFEXP = cmath.exp(1j * k * cn * lR)

    ## setting up the tensor
    xx = CFEXP * (T3XX - ck0 * T2XX - k2 * T1XX) / eps
    yy = CFEXP * (T3YY - ck0 * T2YY - k2 * T1YY) / eps
    zz = CFEXP * (T3ZZ - ck0 * T2ZZ - k2 * T1ZZ) / eps

    xy = CFEXP * (T3XY - ck0 * T2XY - k2 * T1XY) / eps
    xz = CFEXP * (T3XZ - ck0 * T2XZ - k2 * T1XZ) / eps

    yz = CFEXP * (T3YZ - ck0 * T2YZ - k2 * T1YZ) / eps

    yx = xy
    zx = xz
    zy = yz

    G = np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]], dtype=np.complex64)

    return G


def G0_Hp(R1, R2, wavelength, eps):
    """
    R1: dipole position
    R2: evaluation position
    """
    # eps: environment index
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] - R1[2]
    lR2 = Dx**2 + Dy**2 + Dz**2

    k0 = 2 * np.pi / wavelength
    k02n = cmath.sqrt(eps) * k0**2
    # -----------------------------------------------------------------
    T2XY = Dz / lR2
    T3XY = Dz / lR2**1.5
    # -----------------------------------------------------------------
    T2XZ = -Dy / lR2
    T3XZ = -Dy / lR2**1.5
    # -----------------------------------------------------------------
    T2YZ = Dx / lR2
    T3YZ = Dx / lR2**1.5
    # -----------------------------------------------------------------
    CFEXP = -1 * cmath.exp(1j * k0 * cmath.sqrt(eps) * math.sqrt(lR2))

    xx = 0
    yy = 0
    zz = 0

    xy = CFEXP * (1j * k0 * T3XY + k02n * T2XY)
    xz = CFEXP * (1j * k0 * T3XZ + k02n * T2XZ)
    yz = CFEXP * (1j * k0 * T3YZ + k02n * T2YZ)

    yx = -xy
    zx = -xz
    zy = -yz

    G = np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]], dtype=np.complex64)

    return G


def G0_Hm(R1, R2, wavelength, eps):
    return G0_Ep(R1, R2, wavelength, eps)


def G0_Em(R1, R2, wavelength, eps):
    return -1 * np.array(G0_Hp(R1, R2, wavelength, eps))


# =============================================================================
# Quadrupole propagator
# =============================================================================


## --- free space propagator electric quadrupole
def G0_Eqe(R1, R2, wavelength, eps):
    """
    R1: dipole position
    R2: evaluation position
    """
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] - R1[2]
    lR = math.sqrt(Dx**2 + Dy**2 + Dz**2)

    k = 2 * np.pi / wavelength
    cn = cmath.sqrt(eps)
    ck0 = 1j * k * cn
    k2 = k * k * eps

    r3 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 3)
    r25 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 2.5)
    r2 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 2.0)
    r15 = math.pow((Dx * Dx + Dy * Dy + Dz * Dz), 1.5)

    #!C-------------------------------------------------------------------
    T1XX = -1 * (Dy * Dy + Dz * Dz) / r15
    T2XX = (3 * Dx * Dx - 3 * Dy * Dy - 3 * Dz * Dz) / r2
    T3XX = (9 * Dx * Dx - 6 * Dy * Dy - 6 * Dz * Dz) / r25
    T4XX = (9 * Dx * Dx - 6 * Dy * Dy - 6 * Dz * Dz) / r3
    #!C-------------------------------------------------------------------
    T1XY = Dx * Dy / r15
    T2XY = 6 * Dx * Dy / r2
    T3XY = 15 * Dx * Dy / r25
    T4XY = 15 * Dx * Dy / r3
    #!C-------------------------------------------------------------------
    T1XZ = Dx * Dz / r15
    T2XZ = 6 * Dx * Dz / r2
    T3XZ = 15 * Dx * Dz / r25
    T4XZ = 15 * Dx * Dz / r3
    #!C-------------------------------------------------------------------
    T1YY = -(Dx * Dx + Dz * Dz) / r15
    T2YY = (3 * Dy * Dy - 3 * Dx * Dx - 3 * Dz * Dz) / r2
    T3YY = (9 * Dy * Dy - 6 * Dx * Dx - 6 * Dz * Dz) / r25
    T4YY = (9 * Dy * Dy - 6 * Dx * Dx - 6 * Dz * Dz) / r3
    #!C-------------------------------------------------------------------
    T1YZ = Dy * Dz / r15
    T2YZ = 6 * Dy * Dz / r2
    T3YZ = 15 * Dy * Dz / r25
    T4YZ = 15 * Dy * Dz / r3
    #!C------------------------------------------------------------------
    T1ZZ = -(Dx * Dx + Dy * Dy) / r15
    T2ZZ = (3 * Dz * Dz - 3 * Dx * Dx - 3 * Dy * Dy) / r2
    T3ZZ = (9 * Dz * Dz - 6 * Dx * Dx - 6 * Dy * Dy) / r25
    T4ZZ = (9 * Dz * Dz - 6 * Dx * Dx - 6 * Dy * Dy) / r3

    CFEXP = cmath.exp(1j * k * cn * lR)

    ## setting up the tensor
    xx = CFEXP * (T4XX - ck0 * T3XX - k2 * T2XX + ck0 * k2 * T1XX) / (6 * eps)
    yy = CFEXP * (T4YY - ck0 * T3YY - k2 * T2YY + ck0 * k2 * T1YY) / (6 * eps)
    zz = CFEXP * (T4ZZ - ck0 * T3ZZ - k2 * T2ZZ + ck0 * k2 * T1ZZ) / (6 * eps)

    xy = CFEXP * (T4XY - ck0 * T3XY - k2 * T2XY + ck0 * k2 * T1XY) / (6 * eps)
    xz = CFEXP * (T4XZ - ck0 * T3XZ - k2 * T2XZ + ck0 * k2 * T1XZ) / (6 * eps)

    yz = CFEXP * (T4YZ - ck0 * T3YZ - k2 * T2YZ + ck0 * k2 * T1YZ) / (6 * eps)

    yx = xy
    zx = xz
    zy = yz

    G = np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]], dtype=np.complex64)

    return G


def G0_Hqe(R1, R2, wavelength, eps):
    """
    R1: dipole position
    R2: evaluation position
    """
    # eps: environment index
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] - R1[2]
    lR2 = Dx**2 + Dy**2 + Dz**2

    k0 = 2 * np.pi / wavelength
    k02n = cmath.sqrt(eps) * k0**2
    k2 = eps * k0**2
    # -----------------------------------------------------------------
    T2XY = -Dz / lR2
    T3XY = -Dz / lR2**1.5
    T4XY = -Dz / lR2**2
    # -----------------------------------------------------------------
    T2XZ = Dy / lR2
    T3XZ = Dy / lR2**1.5
    T4XZ = Dy / lR2**2
    # -----------------------------------------------------------------
    T2YZ = -Dx / lR2
    T3YZ = -Dx / lR2**1.5
    T4YZ = -Dx / lR2**2
    # -----------------------------------------------------------------
    CFEXP = cmath.exp(1j * k0 * cmath.sqrt(eps) * math.sqrt(lR2))

    xx = 0
    yy = 0
    zz = 0

    xy = CFEXP * (-1j * k0 * k2 * T2XY + 3 * k02n * T3XY + 3j * k0 * T4XY) / 6
    xz = CFEXP * (-1j * k0 * k2 * T2XZ + 3 * k02n * T3XZ + 3j * k0 * T4XZ) / 6
    yz = CFEXP * (-1j * k0 * k2 * T2YZ + 3 * k02n * T3YZ + 3j * k0 * T4YZ) / 6

    yx = -xy
    zx = -xz
    zy = -yz

    G = np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]], dtype=np.complex64)

    return G


def G0_Hqm(R1, R2, wavelength, eps):
    return G0_Eqe(R1, R2, wavelength, eps)


def G0_Eqm(R1, R2, wavelength, eps):
    return -1 * G0_Hqe(R1, R2, wavelength, eps)


# %%

if __name__ == "__main__":
    import numpy as np
    import matplotlib.pyplot as plt
    import copy

    from pyGDM2 import tools

    func_vec = G0_Ep_vec
    func_seq = G0_Ep

    # func_vec = G0_Hp_vec
    # func_seq = G0_Hp

    # test:
    r1 = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    r2 = np.array([[50, 0, 20], [100, 30, 10], [500, 0, 0]])

    g0 = func_vec(r1, r2, wavelength=500, eps_env=1)
    print(g0.shape)

    ##%%

    eps_env = 1
    wl = 500

    DX, NX = 300, 201
    Z0 = 50
    r_probe = tools.generate_NF_map_XY(-DX, DX, NX, -DX, DX, NX, Z0)
    r0 = np.zeros_like(r_probe)

    t0 = time.time()
    G0_vec1 = func_vec(r0, r_probe, wl, eps_env)
    t1 = time.time()
    print(t1 - t0)

    G0_vec2 = np.zeros([len(r_probe), 3, 3], dtype=np.complex64)
    for i, _r in enumerate(r_probe):
        G0_vec2[i] = func_seq(r0[i], _r, wl, eps_env)
    t2 = time.time()
    print(t2 - t1, (t2 - t1) / (t1 - t0))

    G0_vec1 = G0_vec1.reshape(NX, NX, 3, 3)
    G0_vec2 = G0_vec2.reshape(NX, NX, 3, 3)

    print(np.abs(G0_vec1).mean())
    print((np.abs(G0_vec1) - np.abs(G0_vec2)).mean())

    i1, i2 = 2, 1
    plt.subplot(221)
    plt.title("vectorized")
    plt.imshow(G0_vec1[..., i1, i2].real)
    plt.colorbar()
    plt.subplot(222)
    plt.title("scalar")
    plt.imshow(G0_vec2[..., i1, i2].real)
    plt.colorbar()
    plt.subplot(223)
    plt.imshow(G0_vec1[..., i1, i2].imag)
    plt.colorbar()
    plt.subplot(224)
    plt.imshow(G0_vec2[..., i1, i2].imag)
    plt.colorbar()

    plt.show()
